#!/usr/bin/env ruby

#  1659, Wed  4 Jan 12 (NZDT)  v2.1.0
#
#  write-sum.rb: writes summary file for iatmon
#
#  Copyright 2009-2012, Nevil Brownlee, U Auckland | CAIDA | WAND
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

class NbrHash < Hash  # For port numbers and ip lengths
  def update(str)
    if !self[str]
      self[str] = 1
      return true  # new
    else
      self[str] += 1
      return false
    end
  end

  def val_range
    min = 65535;  max = 1
    self.each do |key, val|
      ikey = key.to_i
      min = ikey if ikey < min
      max = ikey if ikey > max
    end
    [self.size, min, max]
  end

  def mode
    shash = self.sort { |kva, kvb| kvb[1] <=> kva[1] }
    shash[0][0].to_i
  end

  def write(f)
    shash = self.sort { |kva, kvb| kvb[1] <=> kva[1] }
    n = 0
    shash.each do |kv|
      f.printf("  %u %u", kv[0], kv[1])
      n += 1
      break if n == N_ProbePorts
    end
  end

  def write_str(f, n_reqd)
    shash = self.sort { |kva, kvb| kvb[1] <=> kva[1] }
    n = 0
    shash.each do |kv|
      f.print"  #{kv[0]} #{kv[1]}"
      n += 1
      break if n == n_reqd
    end
  end
end  # NbrHash


class VnSummary
  def initialize(version)
    @v = version  # 4 or 6

    # type*group matrics (m)
    @src_counts = Array.new(SrcTypeNames.size)  # rows: S_ types
    @pkt_counts = Array.new(SrcTypeNames.size)
    @byte_counts = Array.new(SrcTypeNames.size)
    SrcTypeNames.each_index do |r|  # cols: IAT_ groups
      @src_counts[r] = Array.new(GroupNames.size, 0)
      @pkt_counts[r] = Array.new(GroupNames.size, 0)
      @byte_counts[r] = Array.new(GroupNames.size, 0.0)
    end

    # type distributions
    @t_pkt = Array.new;  @t_byte = Array.new
    @t_life = Array.new;  @t_rate = Array.new  # Whole life of source
    (0..N_source_types-1).each do
      @t_pkt.push(Distribution.new(1, 5000, 100, DT_LOG, 1.0))
      @t_byte.push(Distribution.new(1000, 2000000, 120, DT_LOG, 1.0))
      @t_life.push(Distribution.new(1.0, MxLifetime, 120, DT_LOG, 1.0))
      @t_rate.push(Distribution.new(0.001, 1000.0, 100, DT_LOG, 1.0))
    end

    # group distributions
    @g_pkt = Array.new;  @g_byte = Array.new
    @g_life = Array.new;  @g_rate = Array.new  # Whole life of source
    (0..N_source_groups-1).each do
      @g_pkt.push(Distribution.new(1, 5000, 100, DT_LOG, 1.0))
      @g_byte.push(Distribution.new(1000, 2000000, 120, DT_LOG, 1.0))
      @g_life.push(Distribution.new(1.0, MxLifetime, 120, DT_LOG, 1.0))
      @g_rate.push(Distribution.new(0.001, 1000.0, 100, DT_LOG, 1.0))
    end

    # sundry info
    @tcp_probe_port_counts = NbrHash.new
    @udp_probe_port_counts = NbrHash.new
    @proto_counts = NbrHash.new
  end

  def count(src_type, src_group, to_pkts, to_bytes, life, probe_port, protocol)
    @src_counts[src_type][src_group] += 1
    @pkt_counts[src_type][src_group] += to_pkts
    @byte_counts[src_type][src_group] += to_bytes/1000000.0

    @t_pkt[src_type].count(to_pkts)
    @t_byte[src_type].count(to_bytes)
    @t_life[src_type].count(life)
    rate = to_pkts < MinPkts ? 0 : (to_pkts - 1)/life 
    @t_rate[src_type].count(rate)

    @g_pkt[src_group].count(to_pkts)
    @g_byte[src_group].count(to_bytes)
    @g_life[src_group].count(life)
    rate = to_pkts < MinPkts ? 0 : (to_pkts - 1)/life
    @g_rate[src_group].count(rate)

    if !probe_port.nil?
        # 'probe' means 'all pkts sent to same port'
        # even if source sends only one pkt!  28 Apr 11
      if src_type == S_tcp_probe
        @tcp_probe_port_counts.update(probe_port)
      elsif src_type == S_udp_probe
        @udp_probe_port_counts.update(probe_port)
      end
    end
    @proto_counts.update(protocol)
  end

  def write_matrix(section_head, m, fmt)
    $osf.print section_head + "\n"
    SrcTypeNames.each_index do |r|  # rows (types)
      m[r].each_index do |c|  # cols (groups)
        $osf.printf(fmt, m[r][c])
      end
      $osf.print "\n"
    end
    $osf.print "\n"
  end

  def write_distrib(section_head, d, unit, x_dd)
    $osf.print section_head + "\n#{unit} "
    d[0].write_x($osf, x_dd)  # Nbr of decimal digits for x values
    d.each_index do |st|
      $osf.print "#{st} "  # type/group number
      d[st].write_y($osf, 0)  # Counts for type/group
    end
    $osf.print "\n"
  end

  def write(way_counts)
    write_matrix("\##{@v}msrc (Source Counts)", @src_counts, " %12d")
    write_matrix("\##{@v}mpkt (Source Packets)", @pkt_counts, " %12d")
    write_matrix("\##{@v}mmbyte (Source Volume)", @byte_counts, " %12.6f")

    write_distrib("\##{@v}dpkt (Type Packets)", @t_pkt, 'p', 1)
    write_distrib("\##{@v}dbyte (Type Volume)", @t_byte, 'B', 1)
    write_distrib("\##{@v}dlife (Type Lifetimes)", @t_life, 's', 4)
    write_distrib("\##{@v}drate (Type Rate)", @t_rate, 'p/s', 4)

    write_distrib("\##{@v}dpkt (Group Packets)", @g_pkt, 'p', 1)
    write_distrib("\##{@v}dbyte (Group Volume)", @g_byte, 'B', 1)
    write_distrib("\##{@v}dlife (Group Lifetimes)", @g_life, 's', 4)
    write_distrib("\##{@v}drate (Group Rate)", @g_rate, 'p/s', 4)
    
    $osf.print "\##{@v}tprobe_ports\n#{N_ProbePorts}"
    @tcp_probe_port_counts.write($osf)
    $osf.print "\n\n"

    $osf.print "\##{@v}uprobe_ports\n#{N_ProbePorts}"
    @udp_probe_port_counts.write($osf)
    $osf.print "\n\n"

    $osf.print "\##{@v}protocols\n#{@proto_counts.size}"
    @proto_counts.write($osf)
    $osf.print "\n\n"

    $osf.print "\##{@v}way_counts (S/p//B for 1,2,0.5-way)\n"
    $osf.print "#{way_counts[0]} #{way_counts[1]} #{way_counts[2]}  " +
      "#{way_counts[3]} #{way_counts[4]} #{way_counts[5]}  " +
      "#{way_counts[6]} #{way_counts[7]} #{way_counts[8]}\n\n"
  end
end


def write_sum_file(fname, start_ts, now_ts)
  $osf = File.new(fname, 'w')
  $osf.print "#version\n"
  $osf.print "#{FormatVer} #{OWTmonVer} #{Meter_ver}\n\n"

  $osf.print "#meter_info\n"
  $osf.print "#{MeterLoc} #{utc_gts(start_ts)}" +
    " #{Trace_sz} #{MinPkts} #{TwoWay ? 'two-way' : ''}\n\n"

  v4_info = VnSummary.new(4)
  if LocalV6
    v6_info = VnSummary.new(6);  v6_top32_counts = NbrHash.new
  end

  msg = $owt.each_source do |s|  # look at final attribs for all streams
    src_type, probe_port, to_pkts, total_lifetime, tot_count, version,
      too_few_pkts, protocol, src_group, flags_mode, keys_mode,
      v6_addrs_s32, to_bytes  = s
    if version == 4
      v4_info.count(src_type, src_group, to_pkts, to_bytes, total_lifetime,
        probe_port, protocol)
    elsif version == 6
      v6_info.count(src_type, src_group, to_pkts, to_bytes, total_lifetime,
        probe_port, protocol)
      if !v6_addrs_s32.nil?
        v6_top32_counts.update(sprintf("%08x", v6_addrs_s32))
      end
    else
      print "??? IP version = >#{version}<, #{version.class}\n"
    end
  end

  v4_way_counts, v6_way_counts, erf_sources,
     half_src_addrs, half_src_counts, st_distribs =
     $owt.c_addr_distribs  # Must call this *after* .each_source{}!

  if !msg.nil?  # #Message at end of streams loop
    $owt.log_msg(0, msg)  # Write to log
    $stdout.flush if $clog_level > 0  # and console
  end

  ua_sources, ua_totpkts, ua_pkts, ua_src_addr, ua_dst_addr, ua_ttl,
  ua_proto, ua_gt1prot, ua_dst_port_keys,ua_dst_port_counts,
  ua_tcpkey_keys,ua_tcpkey_counts  = $owt.unanalysed

  v4_info.write(v4_way_counts)

  if LocalV6
    v6_info.write(v6_way_counts)
    n_v6_slash32s = 20  # Only print first 20
    $osf.print "#v6addrs (first 32 bits)\n#{n_v6_slash32s}"
    v6_top32_counts.write_str($osf, n_v6_slash32s)
    $osf.print "\n\n"
  end
 
  if ua_sources != 0
    $osf.print "#unanalysed\n"
    $osf.print "#unanalysed:  #{ua_sources} sources," +
      " #{ua_totpkts} pkts not analysed\n"
    $osf.print "pkts ";  ua_pkts.each { |c| $osf.print " #{c}" }
    $osf.print "\n"
    $osf.print "src_addr ";  ua_src_addr.each { |c| $osf.print " #{c}" }
    $osf.print "\n"
    $osf.print "dst_addr ";  ua_dst_addr.each { |c| $osf.print " #{c}" }
    $osf.print "\n"
    $osf.print "ttl ";  ua_ttl.each { |c| $osf.print " #{c}" }
    $osf.print "\n"
    $osf.print "proto ";  ua_proto.each { |c| $osf.print " #{c}" }
    $osf.print "\n"
    if !ua_dst_port_keys.nil?
      $osf.print "gt1prot #{ua_gt1prot}\n"
      $osf.print "dst_port  #{ua_dst_port_keys.size}"
      ua_dst_port_keys.each_index do |x|
        $osf.printf("  %u %u", ua_dst_port_keys[x], ua_dst_port_counts[x])
      end
      $osf.print "\n"
      if !ua_tcpkey_keys.nil?
        $osf.print "tcp_key  #{ua_tcpkey_keys.size}"
        ua_tcpkey_keys.each_index do |x|
          $osf.printf("  %08x %u", ua_tcpkey_keys[x], ua_tcpkey_counts[x])
        end
        $osf.print "\n"
      end
    end
    $osf.print "\n"
  end

  if TwoWay
    ntp = 16
    if !half_src_addrs.nil? && half_src_addrs.size != 0
      $osf.print "#half-way source addresses\n"
      $osf.print "#{half_src_addrs.size}: "
      (0..(ntp-1)).each do |x|
        break  if half_src_addrs[x].nil?
        $osf.print " #{half_src_addrs[x]} (#{half_src_counts[x]})"
      end
      $osf.print "\n\n"
    end

    if !erf_sources.nil?
      $osf.print "#erf_sources (1/2/3 for 1,2,0.5-way)\n"
      $osf.print "#{erf_sources[0]} #{erf_sources[1]} #{erf_sources[2]}  " +
        "#{erf_sources[3]} #{erf_sources[4]} #{erf_sources[5]}  " +
        "#{erf_sources[6]} #{erf_sources[7]} #{erf_sources[8]}\n"
      $osf.print "\n"
    end
  end

  fin_proc_time = Process::times;  fin_proc_date = Time.now
  time_string = sprintf("(minutes): trace_duration=%.2f" +
      ", elapsed=%.3f, user=%.3f, system=%.3f",
    (now_ts - start_ts)/60.0,
    (fin_proc_date.to_i - $start_proc_date.to_i)/60.0,
    (fin_proc_time.utime - $start_proc_time.utime)/60.0,
    (fin_proc_time.stime - $start_proc_time.stime)/60.0 )
  $start_proc_time = Process::times;  $start_proc_date = Time.now
  $osf.print "#times #{time_string}\n"
  if $clog_level > 0
    print "#Summary: times #{time_string}\n";  $stdout.flush
  end

  $osf.print "\n#eof \n"
  $osf.close
end  # write_sum_file()
